/**
 * The program(s) herein may be used  freely
 * for personal use only. It may not be sold, 
 * rented, leased, sublicensed  or  modified 
 * without permission of  the  author.  
 * 
 *           www.carlosbcruz.com
 *           
 * This program is provided "AS IS"  without 
 * warranty of any kind. In no event I  will
 * be liable  for  any  direct  or  indirect 
 * damage.
 */

package com.carlosbcruz.account.control;

import java.awt.Container;
import java.io.File;
import java.util.ArrayList;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;

import com.carlosbcruz.account.dynamicmodel.AccountCheckGroup;
import com.carlosbcruz.account.dynamicmodel.AccountComment;
import com.carlosbcruz.account.dynamicmodel.AccountCurrentAccountSingleton;
import com.carlosbcruz.account.dynamicmodel.AccountUserAccount;
import com.carlosbcruz.account.modelpersistence.AccountFileException;
import com.carlosbcruz.account.modelpersistence.AccountReader;
import com.carlosbcruz.account.util.AccountAbstractActionDecorator;
import com.carlosbcruz.account.util.AccountMessages;
import com.carlosbcruz.account.util.AccountObserver;
import com.carlosbcruz.account.util.AccountProperties;
import com.carlosbcruz.account.util.AccountSubject;

/**
 * Action that works on the import event
 * 
 * @author Carlos Fernando Bella Cruz - <a href="mailto:pessoal@carlosbcruz.com">pessoal@carlosbcruz.com</a>
 */
public class AccountEventControlImportAction implements AccountObserver {

    private Container container;
    private DefaultMutableTreeNode rootTreeNode;
    private DefaultTreeModel treeModel;
    private AccountAbstractActionDecorator accountAction;

    /**
     * Constructor that stores the paramenters and link this class
     * to the event that triggers the import action
     * @param container Upper container to allow a file dialog to be opened correctly
     * @param rootTreeNode The tree GUI
     * @param treeModel The tree GUI model
     * @param accountAction Action that triggers the import event
     */
    AccountEventControlImportAction(Container container,
        DefaultMutableTreeNode rootTreeNode, DefaultTreeModel treeModel,
        AccountAbstractActionDecorator accountAction) {

        this.container = container;
        this.rootTreeNode = rootTreeNode;
        this.treeModel = treeModel;
        this.accountAction = accountAction;

        accountAction.addObserver(this);
    }

    /* (non-Javadoc)
     * @see com.carlosbcruz.account.util.AccountObserver#update(com.carlosbcruz.account.util.AccountSubject)
     */
    public void update(AccountSubject accountSubject) {

        // Prepare to ask the user for the file
        JFileChooser jFileChooser = new JFileChooser();
        jFileChooser.setMultiSelectionEnabled(false);

        ArrayList fileNamesToRead = new ArrayList();

        jFileChooser.setCurrentDirectory(new File(AccountProperties
                .getStringProperty(AccountProperties.LAST_INPUT_DIRECTORY)));

        int returnValue = jFileChooser.showOpenDialog(container);

        String fileName = null;

        if (returnValue == JFileChooser.APPROVE_OPTION) {

            File files = jFileChooser.getSelectedFile();

            fileName = files.getAbsolutePath();

            AccountProperties.setStringProperty(
                AccountProperties.LAST_INPUT_DIRECTORY, files.getParent());
        } else
            return;

        // A file name was retrieved, clear the tree.
        rootTreeNode.removeAllChildren();
        treeModel.reload();

        // Read the file
        try {
            AccountCurrentAccountSingleton
                    .setAllUserAccounts((new AccountReader())
                            .ReadAllUserAccounts(fileName));

        }
        catch(AccountFileException e) {
            JOptionPane.showMessageDialog(container,
                "Account Application Error - Could not retrieve accounts: "
                    + e.getMessage(), "Account Exception",
                JOptionPane.ERROR_MESSAGE);
            return;
        }

        // Fill up the tree

        // Fill up the root tree
        AccountTreeNodeObject rootNode = new AccountTreeNodeObject(
            AccountCurrentAccountSingleton.getAllUserAccounts().getOwner(),
            AccountTreeNodeObject.ROOT_NODE);
        rootTreeNode.setUserObject(rootNode);

        // Fill up the com.carlosbcruz.account.control node
        String accountNodeTitle = AccountMessages
                .getString("accountsTreeNodeName"); //$NON-NLS-1$
        AccountTreeNodeObject accountTreeNode = new AccountTreeNodeObject(
            accountNodeTitle, AccountTreeNodeObject.ACCOUNT_NODE);
        DefaultMutableTreeNode accountTitleNode = new DefaultMutableTreeNode(
            accountTreeNode);
        treeModel.insertNodeInto(accountTitleNode, rootTreeNode, rootTreeNode
                .getChildCount());

        // Follow all the model and fill up the 
        // tree model with com.carlosbcruz.account.control information
        for (int i = 0; i < AccountCurrentAccountSingleton.getAllUserAccounts()
                .getNumberOfUserAccounts(); i++) {
            AccountUserAccount account = AccountCurrentAccountSingleton
                    .getAllUserAccounts().getUserAccount(i);

            // Fill up the Account Id Node
            AccountTreeNodeObject accountIdNode = new AccountTreeNodeObject(
                account.getAccountId(), AccountTreeNodeObject.ACCOUNT_ID);
            DefaultMutableTreeNode accountNode = new DefaultMutableTreeNode(
                accountIdNode);
            treeModel.insertNodeInto(accountNode, accountTitleNode,
                accountTitleNode.getChildCount());

            // Follow all the years and create a tree node
            // for each year.
            int numberOfYears = account.getNumberOfYears();
            for (int j = 0; j < numberOfYears; j++) {

                AccountTreeNodeObject accountYearNode = new AccountTreeNodeObject(
                    String.valueOf(account.getYearByPosition(j)),
                    AccountTreeNodeObject.ACCOUNT_LEAF);
                DefaultMutableTreeNode yearNode = new DefaultMutableTreeNode(
                    accountYearNode);
                treeModel.insertNodeInto(yearNode, accountNode, accountNode
                        .getChildCount());

            }
        }

        // Fill up the Check Node
        String checkName = AccountMessages.getString("checksTreeNodeName"); //$NON-NLS-1$
        AccountTreeNodeObject checkTreeNode = new AccountTreeNodeObject(
            checkName, AccountTreeNodeObject.CHECK_NODE);
        DefaultMutableTreeNode checkTitleNode = new DefaultMutableTreeNode(
            checkTreeNode);

        treeModel.insertNodeInto(checkTitleNode, rootTreeNode, rootTreeNode
                .getChildCount());

        // Follow all the model and fill up the 
        // tree model with check information
        int numberOfCheckGroups = AccountCurrentAccountSingleton
                .getAllUserAccounts().getNumberOfCheckGroups();

        for (int i = 0; i < numberOfCheckGroups; i++) {

            AccountCheckGroup checkGroup = AccountCurrentAccountSingleton
                    .getAllUserAccounts().getCheckGroup(i);

            // Fill up the leaf check node
            AccountTreeNodeObject checkTreeLeafNode = new AccountTreeNodeObject(
                checkGroup.getCheckGroupIdentification(),
                AccountTreeNodeObject.CHECK_LEAF);
            DefaultMutableTreeNode checkNode = new DefaultMutableTreeNode(
                checkTreeLeafNode);
            treeModel.insertNodeInto(checkNode, checkTitleNode, checkTitleNode
                    .getChildCount());

        }

        // Fill up the comment node
        String commentNodeName = AccountMessages
                .getString("commentsTreeNodeName"); //$NON-NLS-1$
        AccountTreeNodeObject commentTreeNode = new AccountTreeNodeObject(
            commentNodeName, AccountTreeNodeObject.COMMENT_NODE);
        DefaultMutableTreeNode commentTitleNode = new DefaultMutableTreeNode(
            commentTreeNode);
        treeModel.insertNodeInto(commentTitleNode, rootTreeNode, rootTreeNode
                .getChildCount());

        // Follow all the model and fill up the 
        // tree model with comment information
        int numberOfComments = AccountCurrentAccountSingleton
                .getAllUserAccounts().getNumberOfComments();

        for (int i = 0; i < numberOfComments; i++) {
            AccountComment comment = AccountCurrentAccountSingleton
                    .getAllUserAccounts().getAccountComment(i);

            // Fill up the comment leaf node
            AccountTreeNodeObject commentTreeLeafNode = new AccountTreeNodeObject(
                comment.getCommentId(), AccountTreeNodeObject.COMMENT_LEAF);
            DefaultMutableTreeNode commentNode = new DefaultMutableTreeNode(
                commentTreeLeafNode);
            treeModel.insertNodeInto(commentNode, commentTitleNode,
                commentTitleNode.getChildCount());

        }
    }
}
