/**
 * The program(s) herein may be used  freely
 * for personal use only. It may not be sold, 
 * rented, leased, sublicensed  or  modified 
 * without permission of  the  author.  
 * 
 *           www.carlosbcruz.com
 *           
 * This program is provided "AS IS"  without 
 * warranty of any kind. In no event I  will
 * be liable  for  any  direct  or  indirect 
 * damage.
 */

package com.carlosbcruz.account.dynamicmodel;

import java.util.Iterator;
import java.util.Vector;

/**
 * Represents an User Account. An user com.carlosbcruz.account.control has a set of transactions.
 * 
 * @author Carlos Fernando Bella Cruz - <a href="mailto:pessoal@carlosbcruz.com">pessoal@carlosbcruz.com</a>
 */
public class AccountUserAccount {

    private String accountId;

    private Vector transactions = new Vector();

    private Vector listeners = new Vector();

    // Used for optimization for getPositionOfFirstTransactions
    private int cashePositionOfFirstTransactionsLastYear = 0;
    private int cashePositionOfFirstTransactionsNumberOfTransaction = 0;
    private boolean positionOfFirstTransactionInvalidation = true;

    // Used for optimization for getNumberOfTransactions
    private int casheNumberOfTransactionsLastYear = 0;
    private int casheNumberOfTransactionsTotalTransactions = 0;
    private boolean numberOfTransactionsInvalidation = true;

    /**
     * @param accountId The accountId to set
     */
    public AccountUserAccount(String accountId) {

        this.accountId = accountId;
    }

    /**
     * Returns the com.carlosbcruz.account.control Id.
     * @return The com.carlosbcruz.account.control id
     */
    public String getAccountId() {

        return accountId;
    }

    /**
     * Sets the com.carlosbcruz.account.control Id.
     * @param accountId The accountId to set
     */
    public void setAccountId(String accountId) {

        this.accountId = accountId;
    }

    /**
     * Inform the total of transactions existent 
     * @return The total of transactions
     */
    public int getNumberOfAllTransactions() {

        return transactions.size();
    }

    /**
     * Return the transaction for a specifc position 
     * @param position The informed position
     * @return The required transaction
     */
    public AccountTransaction getTransaction(int position) {

        return (AccountTransaction)transactions.get(position);
    }

    /**
     * Returns the number of transactions for a specif year.
     * @param year target year
     * @return number of transactions
     */
    public int getNumberOfTransactions(int year) {

        if (!numberOfTransactionsInvalidation
            && year == casheNumberOfTransactionsLastYear)
            return casheNumberOfTransactionsTotalTransactions;

        numberOfTransactionsInvalidation = false;
        casheNumberOfTransactionsLastYear = year;

        // Count how many transactions exists for the specified year
        int count = 0;
        Iterator i = transactions.iterator();
        while (i.hasNext()) {
            AccountTransaction accountTransaction = (AccountTransaction)i
                    .next();

            if (accountTransaction.getYear() == year)
                count++;
        }

        casheNumberOfTransactionsTotalTransactions = count;

        return count;
    }

    /**
     * Returns the position of the first transactions for a specif year.
     * @param year target year
     * @return transactions position
     */
    public int getPositionOfFirstTransactions(int year) {

        if (!positionOfFirstTransactionInvalidation
            && year == cashePositionOfFirstTransactionsLastYear)
            return cashePositionOfFirstTransactionsNumberOfTransaction;

        positionOfFirstTransactionInvalidation = false;
        cashePositionOfFirstTransactionsLastYear = year;

        int position = 0;
        Iterator i = transactions.iterator();
        while (i.hasNext()) {
            AccountTransaction accountTransaction = (AccountTransaction)i
                    .next();

            if (accountTransaction.getYear() == year)
                break;

            position++;
        }

        cashePositionOfFirstTransactionsNumberOfTransaction = position;

        return position;
    }

    /**
     * Seek for the a specific transaction ina relative position into the
     * specified year
     * @param year  target year
     * @param position position inside the year. Zero reffers to the first one.
     * @return The required transaction
     */
    public AccountTransaction getTransaction(int year, int position) {

        int positionOfFirstTransactions = getPositionOfFirstTransactions(year);

        return (AccountTransaction)transactions.get(positionOfFirstTransactions
            + position);
    }

    /**
     * Inform the number of different years existent
     * @return The number of different years
     */
    public int getNumberOfYears() {

        // Count how many years exists
        int previousYear = 0;
        int count = 0;
        Iterator i = transactions.iterator();
        while (i.hasNext()) {
            AccountTransaction accountTransaction = (AccountTransaction)i
                    .next();

            if (accountTransaction.getYear() != previousYear) {
                previousYear = accountTransaction.getYear();
                count++;
            }
        }

        return count;
    }

    /**
     * Inform the correspondig year by position. The first position is zero
     * @param position Position of the year
     * @return The year
     */
    public int getYearByPosition(int position) {

        // Count how many years exists
        int previousYear = 0;
        int count = 0;
        Iterator i = transactions.iterator();
        while (i.hasNext()) {
            AccountTransaction accountTransaction = (AccountTransaction)i
                    .next();

            if (accountTransaction.getYear() != previousYear) {
                previousYear = accountTransaction.getYear();

                // If is the year required return it
                if (count == position)
                    break;

                count++;
            }
        }

        return previousYear;
    }

    /**
     * Insert a transaction at the end
     * @param transaction The transaction to be inserted
     */
    public void addTransaction(AccountTransaction transaction) {

        transactions.add(transaction);

        // Used to get optimization
        positionOfFirstTransactionInvalidation = true;
        numberOfTransactionsInvalidation = true;

        if (listeners.size() != 0) {

            AccountEvent event = new AccountEvent(accountId, transaction
                    .getYear());

            Iterator i = listeners.iterator();
            while (i.hasNext()) {
                AccountModelListener listener = (AccountModelListener)i.next();

                listener.accountModelChanged(event);
            }

        }
    }

    /**
     * Add a listener to the com.carlosbcruz.account.control model
     * @param listener instance that will listen to model change events.
     */
    public void addAccountModelListener(AccountModelListener listener) {

        if (!listeners.contains(listener))
            listeners.add(listener);
    }

    /**
     * Remove a listener to the com.carlosbcruz.account.control model
     * @param listener instance that will listen to model change events.
     */
    public void removeAccountModelListener(AccountModelListener listener) {

        if (listeners.contains(listener))
            listeners.remove(listener);
    }

}
