/**
 * The program(s) herein may be used  freely
 * for personal use only. It may not be sold, 
 * rented, leased, sublicensed  or  modified 
 * without permission of  the  author.  
 * 
 *           www.carlosbcruz.com
 *           
 * This program is provided "AS IS"  without 
 * warranty of any kind. In no event I  will
 * be liable  for  any  direct  or  indirect 
 * damage.
 */

package com.carlosbcruz.account.modelpersistence;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.text.NumberFormat;
import java.text.ParseException;

import com.carlosbcruz.account.dynamicmodel.AccountAllUserAccounts;
import com.carlosbcruz.account.dynamicmodel.AccountCheck;
import com.carlosbcruz.account.dynamicmodel.AccountCheckGroup;
import com.carlosbcruz.account.dynamicmodel.AccountComment;
import com.carlosbcruz.account.dynamicmodel.AccountModelProvider;
import com.carlosbcruz.account.dynamicmodel.AccountTransaction;
import com.carlosbcruz.account.dynamicmodel.AccountUserAccount;
import com.carlosbcruz.account.util.AccountValues;
import com.carlosbcruz.account.util.NumberNotFoundException;

/**
 * Controls the input of information from a file
 * 
 * @author Carlos Fernando Bella Cruz - <a href="mailto:pessoal@carlosbcruz.com">pessoal@carlosbcruz.com</a>
 */
public class AccountReader implements AccountModelProvider {

    private FileReader fileReader;
    private BufferedReader bufferedReader;

    private final String NEW_LINE = System.getProperty("line.separator");
    public static final String FILE_TYPE_VERSION = "ACCOUNTFILE_V1";

    /* (non-Javadoc)
     * @see com.carlosbcruz.account.control.model.AccountModelProvider#ReadAllUserAccounts(java.lang.String)
     */
    public AccountAllUserAccounts ReadAllUserAccounts(String fileName)
        throws AccountFileException {

        try {

            String accountNames[];
            String checkGroupNames[];
            String commentNames[];

            //
            // Open File
            //

            try {

                fileReader = new FileReader(fileName);
                bufferedReader = new BufferedReader(fileReader);

                //
                // Read Header File
                //

                String line = null;

                //
                // Read the type of file. It shoud be of type ACCOUNTFILE_V1
                //
                line = bufferedReader.readLine();
                if (line == null)
                    throw new AccountFileException("The file " + fileName
                        + " is not valid. Missing ACCOUNTFILE_V1 token");

                String owner = null;
                if (line.equals(FILE_TYPE_VERSION)) {
                    //
                    // Valid format. Read the owner
                    //
                    line = bufferedReader.readLine();
                    if (line == null)
                        throw new AccountFileException(
                            "Error in file format - Expected accounts owner");

                    owner = line;

                    //
                    // Valid format. Read the number of accounts
                    //
                    line = bufferedReader.readLine();
                    if (line == null)
                        throw new AccountFileException(
                            "Error in file format - Expected number of accounts");

                    try {
                        int numberOfAccounts = Integer.parseInt(line);

                        //
                        // Read the name of each com.carlosbcruz.account.control. Each com.carlosbcruz.account.control name is numbered
                        // 1-Account1
                        // 2-Account2
                        // etc
                        //
                        accountNames = new String[numberOfAccounts];
                        for (int i = 0; i < numberOfAccounts; i++) {

                            line = bufferedReader.readLine();
                            if (line == null)
                                throw new AccountFileException(
                                    "Error in file format - Expected name of com.carlosbcruz.account.control");

                            if (line.length() <= 2)
                                throw new AccountFileException(
                                    "Error in file format - Expected a valid com.carlosbcruz.account.control name");
                            accountNames[i] = line.substring(2);
                        }

                    }
                    catch(RuntimeException e) {
                        throw new AccountFileException(
                            "Error in file format - Expected a valid number of accounts");
                    }

                    //
                    // Read the number of check groups
                    //
                    line = bufferedReader.readLine();
                    if (line == null)
                        throw new AccountFileException(
                            "Error in file format - Expected number of check groups");

                    try {
                        int numberOfCheckGroups = Integer.parseInt(line);

                        //
                        // Read the name of each check group. Each check group name is numbered
                        // 1-Checkgroup1
                        // 2-Checkgroup1
                        // etc
                        //
                        checkGroupNames = new String[numberOfCheckGroups];
                        for (int i = 0; i < numberOfCheckGroups; i++) {

                            line = bufferedReader.readLine();
                            if (line == null)
                                throw new AccountFileException(
                                    "Error in file format - Expected name of check group");

                            if (line.length() <= 2)
                                throw new AccountFileException(
                                    "Error in file format - Expected a valid check group name");
                            checkGroupNames[i] = line.substring(2);
                        }

                    }
                    catch(RuntimeException e) {
                        throw new AccountFileException(
                            "Error in file format - Expected a valid number of check groups");
                    }

                    //
                    // Read the number of comments
                    //
                    line = bufferedReader.readLine();
                    if (line == null)
                        throw new AccountFileException(
                            "Error in file format - Expected number of comments");

                    try {
                        int numberOfComments = Integer.parseInt(line);

                        //
                        // Read the name of each comment. Each comment name is numbered
                        // 1-Comment1
                        // 2-Comment2
                        // etc
                        //
                        commentNames = new String[numberOfComments];
                        for (int i = 0; i < numberOfComments; i++) {

                            line = bufferedReader.readLine();
                            if (line == null)
                                throw new AccountFileException(
                                    "Error in file format - Expected name of comments");

                            if (line.length() <= 2)
                                throw new AccountFileException(
                                    "Error in file format - Expected a valid comment name");
                            commentNames[i] = line.substring(2);
                        }

                    }
                    catch(RuntimeException e) {
                        throw new AccountFileException(
                            "Error in file format - Expected a valid number of comments");
                    }

                } else
                    throw new AccountFileException(
                        "Error in File Format Version not Suported");

                bufferedReader.close();
                fileReader.close();

                //
                // Create the all com.carlosbcruz.account.control information
                //
                AccountAllUserAccounts allAccounts = new AccountAllUserAccounts();
                allAccounts.setOwner(owner);

                //
                // Read all the transactions of each com.carlosbcruz.account.control
                //
                for (int i = 0; i < accountNames.length; i++)
                    allAccounts.addUserAccount(ReadAccount(fileName,
                        accountNames[i]));

                //
                // Read all the checks from the check groups
                //
                for (int i = 0; i < checkGroupNames.length; i++)
                    allAccounts.addCheckGroup(ReadChecks(fileName,
                        checkGroupNames[i]));

                //
                // Read all the comments
                //
                for (int i = 0; i < commentNames.length; i++)
                    allAccounts.addComment(ReadComments(fileName,
                        commentNames[i]));

                return allAccounts;

            }
            finally {
                bufferedReader.close();
                fileReader.close();
            }

        }
        catch(FileNotFoundException fnf) {
            throw new AccountFileException("File not found error: "
                + fnf.getMessage());
        }
        catch(IOException ioe) {
            throw new AccountFileException("File i/o  error: "
                + ioe.getMessage());
        }
    }

    /**
     * Method ReadAccount. Read all the transactions of a specific com.carlosbcruz.account.control
     * @param fileName The name of the files to retrieve the information
     * @param accountIdentifier The com.carlosbcruz.account.control name to be readed
     * @return The user Account
     * @throws AccountFileException
     * @throws FileNotFoundException
     * @throws IOException
     */
    private AccountUserAccount ReadAccount(String fileName,
        String accountIdentifier) throws AccountFileException,
        FileNotFoundException, IOException {

        //
        // Open File
        //

        try {

            fileReader = new FileReader(fileName);
            bufferedReader = new BufferedReader(fileReader);

            //
            // Seek the com.carlosbcruz.account.control identifier into the Account file
            //
            String line = null;
            do
                line = bufferedReader.readLine();
            while (line != null && !line.equals(accountIdentifier));
            if (line == null)
                throw new AccountFileException(
                    "The com.carlosbcruz.account.control identifier:"
                        + accountIdentifier + " was not found");

            // Create the Account that will be returned
            AccountUserAccount account = new AccountUserAccount(
                accountIdentifier);

            float resourcesTotal = 0, savingsTotal = 0;

            //
            // Read a transaction
            //
            line = bufferedReader.readLine();
            if (line.equals("FIM " + accountIdentifier))
                throw new AccountFileException(
                    "No Transaction for the Account: " + accountIdentifier);

            int year = 0;
            do {

                if (line.length() < 5) {
                    //
                    // Read Year
                    //
                    try {
                        year = Integer.parseInt(line);
                    }
                    catch(NumberFormatException e) {
                        throw new AccountFileException(
                            "Could not parse year. Line: \"" + line + "\"");
                    }

                } else {

                    //
                    // Read Day and Month
                    //

                    String dayReaded = line.substring(0, 2);
                    String monthReaded = line.substring(3, 5);

                    int day = 0;
                    try {
                        day = Integer.parseInt(dayReaded);
                    }
                    catch(NumberFormatException e) {
                        throw new AccountFileException(
                            "Could not parse day. Line: \"" + line + "\"");
                    }

                    int month = 0;
                    try {
                        month = Integer.parseInt(monthReaded);
                    }
                    catch(NumberFormatException e) {
                        throw new AccountFileException(
                            "Could not parse month. Line: \"" + line + "\"");
                    }

                    //
                    // Read Withdraw information
                    //

                    if (line.length() < 16) {
                        throw new AccountFileException(
                            "Could not parse widthdraw due to line too short. Line: \""
                                + line + "\"");
                    }

                    //
                    // Extract withdraw information from the line
                    //
                    String withdrawReaded = line.substring(5, 16).trim();

                    //
                    // Store the withdraw in a Number variable withdraw
                    //
                    Number withdraw = null;
                    if (withdrawReaded.length() != 0) {

                        //
                        // A withdraw as found. Parse it
                        //
                        NumberFormat formatter = NumberFormat.getInstance();

                        try {
                            withdraw = formatter.parse(withdrawReaded);
                        }
                        catch(ParseException e) {
                            throw new AccountFileException(
                                "Could not parse withdraw. Line: \"" + line
                                    + "\" Withdraw: \"" + withdrawReaded + "\"");
                        }
                    }

                    //
                    // Read Resource
                    //
                    if (line.length() < 27) {
                        throw new AccountFileException(
                            "Could not parse resource due to line too short. Line: \""
                                + line + "\"");
                    }

                    //
                    // Extract resource from line
                    //
                    String resourceReaded = line.substring(16, 27).trim();

                    Number resource = null;
                    if (resourceReaded.length() != 0) {

                        NumberFormat formatter = NumberFormat.getInstance();

                        try {
                            resource = formatter.parse(resourceReaded);

                        }
                        catch(ParseException e) {
                            throw new AccountFileException(
                                "Could not parse resource. Line: \"" + line
                                    + "\" Resource: \"" + resourceReaded + "\"");
                        }
                    }

                    //
                    // Read Savings
                    //
                    Number savings = null;

                    if (line.length() >= 38) {

                        String savingsReaded = line.substring(27, 38).trim();

                        if (savingsReaded.length() != 0) {

                            NumberFormat formatter = NumberFormat.getInstance();

                            try {
                                savings = formatter.parse(savingsReaded);

                            }
                            catch(ParseException e) {
                                throw new AccountFileException(
                                    "Could not parse savings. Line: \"" + line
                                        + "\" Savings: \"" + savingsReaded
                                        + "\"");
                            }
                        }
                    }

                    //
                    // Read Description if exists
                    //
                    String description = null;
                    if (line.length() > 39)
                        description = line.substring(38).trim();

                    //
                    // If withdraw present than it is a transaction that removes chach from the com.carlosbcruz.account.control
                    //
                    AccountTransaction transaction = new AccountTransaction();
                    if (withdraw != null) {

                        //
                        // withdraw from com.carlosbcruz.account.control
                        //
                        transaction.setDay(day);
                        transaction.setMonth(month);
                        transaction.setYear(year);
                        transaction.setValue(withdraw.floatValue());
                        transaction.setDescription(description);
                        transaction.setType(AccountTransaction.WITHDRAW);

                        account.addTransaction(transaction);

                        resourcesTotal -= withdraw.floatValue();

                    } else
                        if (resource == null) {

                            //
                            // It is a transaction that receives a deposit into the savings com.carlosbcruz.account.control.
                            // The deposit amount is in the description
                            //

                            //
                            // Parse the description and get the deposit amount
                            //
                            Number depositValue = getValueFromDescription(description);

                            transaction.setDay(day);
                            transaction.setMonth(month);
                            transaction.setYear(year);
                            transaction.setValue(depositValue.floatValue());
                            transaction
                                    .setType(AccountTransaction.DEPOSIT_SAVINGS_ACCOUNT);
                            transaction.setDescription(description);

                            savingsTotal += depositValue.floatValue();

                            account.addTransaction(transaction);

                        } else
                            if (savings != null) {
                                //
                                // It is a transation that moves cash FROM or TO the
                                // savings com.carlosbcruz.account.control the the current com.carlosbcruz.account.control. 
                                //

                                float amountTransfered = savingsTotal
                                    - savings.floatValue();

                                // Keep the value positive
                                if (amountTransfered < 0)
                                    amountTransfered *= -1;

                                transaction.setDay(day);
                                transaction.setMonth(month);
                                transaction.setYear(year);
                                transaction.setValue(amountTransfered);

                                // If the next savings is less than the total
                                // than resource was removed from the savings
                                if (savings.floatValue() < savingsTotal) {

                                    transaction
                                            .setType(AccountTransaction.MOVE_FROM_SAVINGS_TO_CURRENT_ACCOUNT);
                                    transaction
                                            .setDescription("Move from Savings to Current Account");
                                    resourcesTotal += amountTransfered;
                                    savingsTotal -= amountTransfered;

                                } else {

                                    transaction
                                            .setType(AccountTransaction.MOVE_FROM_CURRENT_ACCOUNT_TO_SAVINGS);
                                    transaction
                                            .setDescription("Move from Current Account to Savings");
                                    resourcesTotal -= amountTransfered;
                                    savingsTotal += amountTransfered;

                                }

                                account.addTransaction(transaction);

                            } else
                                if (resource != null) {

                                    //
                                    // It is a deposit transaction into the current com.carlosbcruz.account.control
                                    //

                                    //
                                    // Parse the description and get the deposit amount
                                    //
                                    Number depositValue = getValueFromDescription(description);

                                    transaction.setDay(day);
                                    transaction.setMonth(month);
                                    transaction.setYear(year);
                                    transaction.setDescription(description);
                                    transaction.setValue(depositValue
                                            .floatValue());
                                    transaction
                                            .setType(AccountTransaction.DEPOSIT_CURRENT_ACCOUNT);

                                    resourcesTotal += depositValue.floatValue();

                                    account.addTransaction(transaction);

                                } else
                                    throw new AccountFileException(
                                        " Could not know the transaction type. Line: \""
                                            + line + "\"");

                    //
                    // Remove the precision errors from the float point operations
                    //

                    // Garantee two digit decimal on float
                    resourcesTotal = AccountValues
                            .getTwoDecimalFloat(resourcesTotal);

                    //
                    // Check the calculated total resource ammount with the
                    // total informed at the file.
                    //
                    if (resource != null
                        && resourcesTotal != resource.floatValue()) {

                        throw new AccountFileException(
                            "Error in resource calculation. Calculated: "
                                + resourcesTotal + " Found: "
                                + resource.floatValue() + " Line: \"" + line
                                + "\"");
                    }

                    //
                    // Remove the precision errors from the float point operations
                    //

                    // Garantee two digit decimal on float
                    savingsTotal = AccountValues
                            .getTwoDecimalFloat(savingsTotal);

                    //
                    // Check the calculated total savings ammount with the
                    // total informed at the file.
                    //
                    if (savings != null && savingsTotal != savings.floatValue()) {

                        System.out
                                .println("Error in savings calculation. Calculated: "
                                    + savingsTotal
                                    + " Found: "
                                    + savings.floatValue()
                                    + " Line: \""
                                    + line
                                    + "\"");
                        throw new AccountFileException(
                            "Error in savings calculation. Calculated: "
                                + savingsTotal + " Found: "
                                + savings.floatValue() + " Line: \"" + line
                                + "\"");
                    }

                }

                line = bufferedReader.readLine();
                if (line == null)
                    throw new AccountFileException(
                        "Error, unexpected end-of-file while reading com.carlosbcruz.account.control"
                            + accountIdentifier);

            }
            while (!line.equals("FIM " + accountIdentifier));

            bufferedReader.close();
            fileReader.close();

            return account;

        }
        finally {

            bufferedReader.close();
            fileReader.close();

        }
    }

    /**
     * Method getValueFromDescription. Parse a value inside the description of a transaction.
     * @param description The total string representing the description
     * @return The number founded inside the description
     * @throws AccountFileException
     */
    private Number getValueFromDescription(String description)
        throws AccountFileException {

        Number depositValue = null;
        try {
            depositValue = AccountValues.getNumberFromText(description);
        }
        catch(NumberNotFoundException e) {
            throw new AccountFileException(
                "Error trying to parse value from description");
        }
        return depositValue;
    }

    /**
     * Method ReadChecks. Read the checks from a check group
     * @param fileName The name of the file to read the check group
     * @param checkGroupIdentifier The check group identification
     * @return The Account Check Instantiated with the Check Group Requested
     * @throws AccountFileException
     * @throws FileNotFoundException
     * @throws IOException
     */
    private AccountCheckGroup ReadChecks(String fileName,
        String checkGroupIdentifier) throws AccountFileException,
        FileNotFoundException, IOException {

        //
        // Open File
        //

        try {

            fileReader = new FileReader(fileName);
            bufferedReader = new BufferedReader(fileReader);

            //
            // Seek the check group identifier into the Account file
            //
            String line = null;
            do
                line = bufferedReader.readLine();
            while (line != null && !line.equals(checkGroupIdentifier));
            if (line == null)
                throw new AccountFileException("The check group identifier:"
                    + checkGroupIdentifier + " was not found");

            //
            // Read a check
            //
            line = bufferedReader.readLine();
            if (line.equals("FIM " + checkGroupIdentifier))
                throw new AccountFileException(
                    "No Checks found for the Check Group: "
                        + checkGroupIdentifier);

            AccountCheckGroup checkGroup = new AccountCheckGroup();
            checkGroup.setCheckGroupIdentification(checkGroupIdentifier);

            do {

                //
                // Read and parse the check sequence number
                //
                if (line.length() <= 4)
                    throw new AccountFileException(
                        "Not a valid check line information. Line: \"" + line
                            + "\"");

                String checkSequence = line.substring(0, 4).trim();
                NumberFormat formatter = NumberFormat.getInstance();

                Number checkSequenceNumber = null;
                try {
                    checkSequenceNumber = formatter.parse(checkSequence);
                }
                catch(ParseException e) {
                    throw new AccountFileException(
                        "Could not parse check sequence number. Line: \""
                            + line + "\"");
                }

                //
                // Read the check identification
                //
                if (line.length() <= 12)
                    throw new AccountFileException(
                        "Not a valid check line information. Line: \"" + line
                            + "\"");

                String checkIdentification = line.substring(6, 12).trim();

                //
                // Read the check value
                //
                if (line.length() <= 22)
                    throw new AccountFileException(
                        "Not a valid check line information. Line: \"" + line
                            + "\"");

                String checkValueReaded = line.substring(12, 22).trim();
                formatter = NumberFormat.getInstance();

                Number checkValue = null;
                try {
                    checkValue = formatter.parse(checkValueReaded);
                }
                catch(ParseException e) {
                    throw new AccountFileException(
                        "Could not parse check value. Line: \"" + line + "\"");
                }

                //
                // Read the check description
                //
                String checkDescription;
                if (line.length() >= 63)
                    checkDescription = line.substring(23, 63).trim();
                else
                    checkDescription = line.substring(23).trim();

                //
                // Read the check status
                //
                String checkStatus = null;
                if (line.length() >= 63)
                    checkStatus = line.substring(63).trim();

                AccountCheck check = new AccountCheck();
                check.setCheckSequence(checkSequenceNumber.intValue());
                check.setCheckIdentification(checkIdentification);
                check.setValue(checkValue.floatValue());
                check.setDescription(checkDescription);
                check.setStatus(checkStatus);

                checkGroup.addCheck(check);

                line = bufferedReader.readLine();
                if (line == null)
                    throw new AccountFileException(
                        "Error, unexpected end-of-file while reading check group"
                            + checkGroupIdentifier);

            }
            while (!line.equals("FIM " + checkGroupIdentifier));

            bufferedReader.close();
            fileReader.close();

            return checkGroup;

        }
        finally {
            bufferedReader.close();
            fileReader.close();

        }

    }

    /**
     * Method ReadComments. Read a specific comment 
     * @param fileName The name of the file to read the comment
     * @param commentIdentifier The comment identification
     * @return The comment class instantiated with the comment
     * @throws AccountFileException
     * @throws FileNotFoundException
     * @throws IOException
     */
    private AccountComment ReadComments(String fileName,
        String commentIdentifier) throws AccountFileException,
        FileNotFoundException, IOException {

        //
        // Open File
        //

        try {

            fileReader = new FileReader(fileName);
            bufferedReader = new BufferedReader(fileReader);

            //
            // Seek for the commentIdentifiere
            //
            String line = null;
            do
                line = bufferedReader.readLine();
            while (line != null && !line.equals(commentIdentifier));
            if (line == null)
                throw new AccountFileException("The comment identifier:"
                    + commentIdentifier + " was not found");

            //
            // Read the comment
            //
            line = bufferedReader.readLine();
            if (line.equals("FIM " + commentIdentifier))
                throw new AccountFileException(
                    "No Comments found for the comments: " + commentIdentifier);

            StringBuffer comment = new StringBuffer();

            do {

                //
                // Read a comment line
                //
                comment.append(line + NEW_LINE);

                line = bufferedReader.readLine();
                if (line == null)
                    throw new AccountFileException(
                        "Error, unexpected end-of-file while reading comment"
                            + commentIdentifier);

            }
            while (!line.equals("FIM " + commentIdentifier));

            bufferedReader.close();
            fileReader.close();

            //
            // Store the comment
            //
            AccountComment accountComment = new AccountComment();
            accountComment.setCommentId(commentIdentifier);
            accountComment.setComments(comment);

            return accountComment;

        }
        finally {
            bufferedReader.close();
            fileReader.close();

        }

    }
}