/**
 * The program(s) herein may be used  freely
 * for personal use only. It may not be sold, 
 * rented, leased, sublicensed  or  modified 
 * without permission of  the  author.  
 * 
 *           www.carlosbcruz.com
 *           
 * This program is provided "AS IS"  without 
 * warranty of any kind. In no event I  will
 * be liable  for  any  direct  or  indirect 
 * damage.
 */

package com.carlosbcruz.account.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Properties;
import java.util.Vector;

/**
 * General Properties for the Account Application
 * 
 * @author Carlos Fernando Bella Cruz - <a href="mailto:pessoal@carlosbcruz.com">pessoal@carlosbcruz.com</a>
 */
public class AccountProperties {

    private static Properties properties = new Properties();

    private static boolean propertiesFileExists = false;

    private static String PROPERTIES_FILE_NAME = "Account.properties";
    private static String PROPERTIES_FILE_HEADER = "Account Properties";

    private static String GROUP_PREFIX = "Group";
    public static String LAST_INPUT_DIRECTORY = "LastDirectory";
    public static String QUICK_DESCRIPTION_GROUP = "QuickDescription";
    public static String BANK_TYPE_DESCRIPTION_GROUP = "BankTypeDescription";

    static {

        File inputPropertiesFile = new File(PROPERTIES_FILE_NAME);
        FileInputStream inputStream = null;
        try {
            inputStream = new FileInputStream(inputPropertiesFile);
            propertiesFileExists = true;
            try {
                properties.load(inputStream);

            }
            catch(IOException e1) {

                e1.printStackTrace();
            }
        }
        catch(FileNotFoundException e) {

            // Do nothing
        }

    }

    /**
     * Get the application default value for each key
     * @param propertyKey The parameter key
     * @return the value
     */
    private static String getDefaultStringProperty(String propertyKey) {

        if (propertyKey.equals(LAST_INPUT_DIRECTORY))
            return "C:/";

        return null;
    }

    /**
     * Get the property to a key. A default value is returned if not found
     * @param propertyKey The parameter key
     * @return the value
     */
    public static String getStringProperty(String propertyKey) {

        String property = properties.getProperty(propertyKey);

        if (property == null)
            return getDefaultStringProperty(propertyKey);

        return property;
    }

    /**
     * Return all the keys that belongs to the property group
     * @param propertyKey property key
     * @return all the keys for the property group informed
     */
    private static Vector getGroupKeysProperty(String propertyKey) {

        Vector keys = new Vector();

        String prefix = GROUP_PREFIX + propertyKey;
        Enumeration e = properties.keys();
        while (e.hasMoreElements()) {
            String key = (String)e.nextElement();

            if (key.length() > prefix.length()) {

                if (key.substring(0, prefix.length()).equals(prefix))
                    keys.add(key);
            }
        }

        return keys;
    }

    /**
     * Remove all the properties that belongs to a group
     * @param propertyKey The group to be removed
     */
    public static void removeAllGroupKeysProperty(String propertyKey) {

        Vector keys = getGroupKeysProperty(propertyKey);

        Enumeration e = keys.elements();
        while (e.hasMoreElements()) {

            String key = (String)e.nextElement();
            properties.remove(key);
        }
    }

    /**
     * Remove a property from a group informing the position
     * @param propertyKey The group key identification
     * @param position Position to be removed
     */
    public static void removeGroupKeyPropertyByPosition(String propertyKey,
        int position) {

        Vector keys = getGroupKeysProperty(propertyKey);
        String prefix = GROUP_PREFIX + propertyKey;

        properties.remove(prefix + position);

        for (int i = position + 1; i < keys.size(); i++) {
            String valueStored = properties.getProperty(prefix + i);

            properties.remove(prefix + i);
            properties.setProperty(prefix + (i - 1), valueStored);
        }
    }

    /**
     * Insert a property at the end of a group
     * @param propertyKey Property Group
     * @param value The value
     */
    public static void addGroupProperty(String propertyKey, String value) {

        Vector keys = getGroupKeysProperty(propertyKey);

        properties.setProperty(GROUP_PREFIX + propertyKey + (keys.size()),
            value);

    }

    /**
     * Return a group of properties specifying the group property key
     * @param propertyKey group property key 
     * @return The vector with all the group contents
     */
    public static Vector getGroupProperty(String propertyKey) {

        Vector returnValues = new Vector();
        Vector keys = getGroupKeysProperty(propertyKey);

        Enumeration e = keys.elements();
        while (e.hasMoreElements()) {
            String key = (String)e.nextElement();

            String value = getStringProperty(key);
            returnValues.add(value);
        }

        return returnValues;
    }

    /**
     * Return specific groups property element by its index
     * @param propertyKey group property key
     * @param index the property position
     * @return The property requested
     */
    public static String getGroupProperty(String propertyKey, int index) {

        return properties.getProperty(GROUP_PREFIX + propertyKey + index);
    }

    /**
     * Set an Account property
     * @param propertyKey the Key
     * @param propertyValue the Value
     */
    public static void setStringProperty(String propertyKey,
        String propertyValue) {

        properties.setProperty(propertyKey, propertyValue);
    }

    /**
     * Save the properties in the application file
     */
    public static void saveProperties() {

        File outputPropertiesFile = new File(PROPERTIES_FILE_NAME);

        try {

            outputPropertiesFile.createNewFile();

        }
        catch(IOException e1) {

            e1.printStackTrace();
            return;
        }

        FileOutputStream outputStream = null;
        try {

            outputStream = new FileOutputStream(outputPropertiesFile);
            propertiesFileExists = true;

        }
        catch(FileNotFoundException e) {
            // Do nothing
        }

        try {

            properties.store(outputStream, PROPERTIES_FILE_HEADER);

        }
        catch(IOException e2) {

            e2.printStackTrace();
            return;
        }

    }
}
